/*
 * Copyright © 2013 Ian D. Romanick
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#include <math.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <stdio.h>
#include <GL/glew.h>
#include <SDL.h>
#include <glu3.h>
#include <assert.h>
#include <unistd.h>

#include "texture_load.h"
#include "framebuffer.h"

#define BUFFER_OFFSET(i) ((char *)NULL + (i))
#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

static SDL_Surface *my_surf = NULL;
static bool done = false;
static GLint w;
static GLint h;
static GLuint fbo;

/**
 * images[0] is the original image, and images[1] is the de-noised image.
 */
static GLuint images[2];

static GLuint difference_kernel;
static GLuint alt_weight_kernel;
static GLuint alt_weight_kernel00;
static GLuint normalize_kernel;
static GLuint normalize_color_kernel;
static GLuint pass[4];
static bool draw = true;

/* Is the de-noised image available?
 */
static bool denoised = false;

enum images {
	image_original,
	image_denoised,
	image_difference
};

enum images what_to_show = image_original;

static void nlmeans2(GLuint src_image, GLuint dst_image);

static void nlmeans3(GLuint src_image, GLuint dst_image);

static void
Redisplay()
{
	if (!draw)
		return;

	/* If the user has requested any display that requires the denoised
	 * image data, but that data is not available, generate it.
	 */
	if (what_to_show != image_original && !denoised) {
		nlmeans3(images[0], images[1]);
		denoised = true;
	}

	switch (what_to_show) {
	case image_original:
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
		glViewport(0, 0, my_surf->w, my_surf->h);

		glBindFramebuffer(GL_READ_FRAMEBUFFER, fbo);
		glFramebufferTexture2D(GL_READ_FRAMEBUFFER,
				       GL_COLOR_ATTACHMENT0,
				       GL_TEXTURE_2D,
				       images[0],
				       0 /* level */);
		assert(glCheckFramebufferStatus(GL_READ_FRAMEBUFFER)
		       == GL_FRAMEBUFFER_COMPLETE);

		glBlitFramebuffer(0, 0,
				  w, h,
				  0, 0,
				  my_surf->w, my_surf->h,
				  GL_COLOR_BUFFER_BIT,
				  GL_LINEAR);
		break;

	case image_denoised:
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
		glViewport(0, 0, my_surf->w, my_surf->h);

		glBindFramebuffer(GL_READ_FRAMEBUFFER, fbo);
		glFramebufferTexture2D(GL_READ_FRAMEBUFFER,
				       GL_COLOR_ATTACHMENT0,
				       GL_TEXTURE_2D,
				       images[1],
				       0 /* level */);
		assert(glCheckFramebufferStatus(GL_READ_FRAMEBUFFER)
		       == GL_FRAMEBUFFER_COMPLETE);

		glBlitFramebuffer(0, 0,
				  w, h,
				  0, 0,
				  my_surf->w, my_surf->h,
				  GL_COLOR_BUFFER_BIT,
				  GL_LINEAR);
		break;

	case image_difference: {
		GLint loc;

		glBindFramebuffer(GL_READ_FRAMEBUFFER, 0);
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
		glViewport(0, 0, my_surf->w, my_surf->h);

		glUseProgram(difference_kernel);
		loc = glGetUniformLocation(difference_kernel, "a");
		assert(loc >= 0);
		glUniform1i(loc, 0);

		loc = glGetUniformLocation(difference_kernel, "b");
		assert(loc >= 0);
		glUniform1i(loc, 1);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, images[1]);

		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, images[0]);

		glDrawArrays(GL_QUADS, 0, 4);

		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, 0);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, 0);
		break;
	}
	}

	SDL_GL_SwapBuffers();
	draw = false;
}

/**
 * Set useful default texture paramters for the currently bound texture.
 */
void
initialize_texture_parameters(void)
{
	/* Sampling outside the texture should always return (0., 0.).
	 * This greatly simplifies the down-sample kernel.
	 */
	static const float border[] = { 0.f,  0.f,  0.f,  0.f };

	glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_WRAP_S,
			GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_WRAP_T,
			GL_CLAMP_TO_BORDER);
	glTexParameterfv(GL_TEXTURE_2D,
			 GL_TEXTURE_BORDER_COLOR,
			 border);

	/* Ensure texture completeness by forcing the min filter to
	 * nearest.
	 */
	glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_MIN_FILTER,
			GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_MAG_FILTER,
			GL_NEAREST);
}


void
generate_framebuffer(struct framebuffer *fb, unsigned w, unsigned h)
{
	/* Generate an RG fp32 framebuffer that will contain the per-pixel
	 * values for (w, p * w).
	 */
	glGenFramebuffers(1, &fb->fbo);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, fb->fbo);

	fb->width = w;
	fb->height = h;

	glGenTextures(ARRAY_SIZE(fb->color_textures),
		      fb->color_textures);

	for (unsigned i = 0; i < ARRAY_SIZE(fb->color_textures); i++) {
		glBindTexture(GL_TEXTURE_2D,
			      fb->color_textures[i]);

		glTexImage2D(GL_TEXTURE_2D, 0 /* level */, GL_RG32F,
			     w, h, 0 /* border */,
			     GL_RG, GL_FLOAT, NULL /* data */);

		initialize_texture_parameters();
	}
}


static void
nlmeans2(GLuint src_image, GLuint dst_image)
{
	const int window_size = 2;

	unsigned width;
	unsigned height;
	struct framebuffer fb;
	GLuint dst_fbo;
	static GLuint q = 0;

	if (q == 0) {
		glGenQueries(1, &q);
	}

	glBindTexture(GL_TEXTURE_2D, src_image);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,
				 0 /* level */,
				 GL_TEXTURE_WIDTH,
				 (GLint *) &width);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,
				 0 /* level */,
				 GL_TEXTURE_HEIGHT,
				 (GLint *) &height);

	generate_framebuffer(&fb, width, height);

	const float texel_width = float(1. / float(width));
	const float texel_height = float(1. / float(height));

	const GLuint p = alt_weight_kernel;
	const GLint loc = glGetUniformLocation(p, "q");
	assert(loc >= 0);

	glBindTexture(GL_TEXTURE_2D,
		      src_image);

	glBindFramebuffer(GL_DRAW_FRAMEBUFFER,
			  fb.fbo);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER,
			       GL_COLOR_ATTACHMENT0,
			       GL_TEXTURE_2D,
			       fb.color_textures[0],
			       0 /* level */);
	assert(glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER)
	       == GL_FRAMEBUFFER_COMPLETE);

	glBeginQuery(GL_TIME_ELAPSED, q);

	glUseProgram(p);

	/* Configure the blend unit to add the results of one pass to the
	 * accumulated results of the previous passes.  Blending is disabled
	 * on the first pass because there is no data from previous passes.
	 */
	glBlendEquationSeparate(GL_FUNC_ADD, GL_FUNC_ADD);
	glBlendFuncSeparate(GL_ONE, GL_ONE, GL_ZERO, GL_ZERO);
	glDisable(GL_BLEND);

	bool first_pass = true;
	for (int y = -window_size; y < window_size; y += 2) {
		for (int x = -window_size; x < window_size; x += 2) {
			if (x == 0 && y == 0) {
				glUseProgram(alt_weight_kernel00);
			} else {
				glUniform2f(loc,
					    texel_width * float(x),
					    texel_height * float(y));
			}

			glDrawArrays(GL_QUADS, 0, 4);

			/* After the first pass, enable blending to accumulate
			 * future results.
			 */
			if (first_pass) {
				glEnable(GL_BLEND);
				first_pass = false;
			}

			if (x == 0 && y == 0) {
				glUseProgram(p);
			}
		}
	}

	/* Re-disable blending.
	 */
	glDisable(GL_BLEND);


	/* Finally normalize the accumulated (w, k * w) data for display.
	 */
	glGenFramebuffers(1, &dst_fbo);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, dst_fbo);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER,
			       GL_COLOR_ATTACHMENT0,
			       GL_TEXTURE_2D,
			       dst_image,
			       0 /* level */);

	
	glUseProgram(normalize_kernel);
	glBindTexture(GL_TEXTURE_2D, fb.color_textures[0]);
	glDrawArrays(GL_QUADS, 0, 4);

	/* Emit timing information for the curious. :)
	 */
	glEndQuery(GL_TIME_ELAPSED);

	GLuint64 result;
	glGetQueryObjectui64v(q,
			      GL_QUERY_RESULT,
			      &result);
	printf("%fms per pixel (%ld.%ldms total)\n",
	       double(result) / (1000000. * width * height),
	       result / 1000000,
	       (result / 10000) % 100);

	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);

}

static void
nlmeans3(GLuint src_image, GLuint dst_image)
{
	const int window_size = 2;

	unsigned width;
	unsigned height;

	GLuint tmp_fbo;
	GLuint dst_fbo;
	GLuint textures[3];

	GLuint q;

	glGenQueries(1, &q);

	glBindTexture(GL_TEXTURE_2D, src_image);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,
				 0 /* level */,
				 GL_TEXTURE_WIDTH,
				 (GLint *) &width);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,
				 0 /* level */,
				 GL_TEXTURE_HEIGHT,
				 (GLint *) &height);

	glGenTextures(ARRAY_SIZE(textures), textures);

	glBindTexture(GL_TEXTURE_2D, textures[0]);
	glTexImage2D(GL_TEXTURE_2D, 0 /* level */, GL_RGB10_A2_EXT,
		     width, height, 0 /* border */,
		     GL_RED, GL_FLOAT, NULL /* data */);
	initialize_texture_parameters();

	glBindTexture(GL_TEXTURE_2D, textures[1]);
	glTexImage2D(GL_TEXTURE_2D, 0 /* level */, GL_RGB10_A2_EXT,
		     width, height, 0 /* border */,
		     GL_RED, GL_FLOAT, NULL /* data */);
	initialize_texture_parameters();

	glBindTexture(GL_TEXTURE_2D, textures[2]);
	glTexImage2D(GL_TEXTURE_2D, 0 /* level */, GL_RGBA16F,
		     width, height, 0 /* border */,
		     GL_RGBA, GL_FLOAT, NULL /* data */);
	initialize_texture_parameters();

	glGenFramebuffers(1, &tmp_fbo);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, tmp_fbo);
	glViewport(0, 0, width, height);

	const float texel_width = float(1. / float(width));
	const float texel_height = float(1. / float(height));

	GLint q_loc[4] = { -1, -1, -1, -1 };
	GLint image_loc[4] = { -1, -1, -1, -1 };
	GLint kernel_loc[4] = { -1, -1, -1, -1 };
	GLint d_loc[4] = { -1, -1, -1, -1 };
	GLint g_loc[4] = { -1, -1, -1, -1 };

	for (unsigned i = 0; i < ARRAY_SIZE(pass); i++) {
		q_loc[i] = glGetUniformLocation(pass[i], "q");
		image_loc[i] = glGetUniformLocation(pass[i], "image");
		d_loc[i] = glGetUniformLocation(pass[i], "d");
		g_loc[i] = glGetUniformLocation(pass[i], "g");
		kernel_loc[i] = glGetUniformLocation(pass[i], "kernel");

		glUseProgram(pass[i]);
		glUniform1i(image_loc[i], 0);
		glUniform1i(d_loc[i], 1);
		glUniform1i(g_loc[i], 2);
		glUniform1i(kernel_loc[i], 1);
	}

	/* "image" in the shaders.
	 */
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,
		      src_image);

	/* "d" and "kernel" in the shaders.
	 */
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D,
		      textures[0]);

	/* "g" in the shaders.
	 */
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D,
		      textures[1]);

	/* "s" in the normalize shader.
	 */
	glActiveTexture(GL_TEXTURE3);
	glBindTexture(GL_TEXTURE_2D,
		      textures[2]);

	/* Configure the blend unit to add the results of one pass to the
	 * accumulated results of the previous passes.  Blending is disabled
	 * on the first pass because there is no data from previous passes.
	 */
	glBlendEquationSeparate(GL_FUNC_ADD, GL_FUNC_ADD);
	glBlendFuncSeparate(GL_ONE, GL_ONE, GL_ONE, GL_ONE);
	glDisable(GL_BLEND);

	glBeginQuery(GL_TIME_ELAPSED, q);
	for (int y = -window_size; y <= window_size; y++) {
		for (int x = 0; x <= window_size; x += 3) {
			/* Begin pass 1: emit the raw distances between pixels
			 * at p and p+q.
			 */
			glUseProgram(pass[0]);
			glUniform2f(q_loc[0],
				    texel_width * float(x),
				    texel_height * float(y));

			glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER,
					       GL_COLOR_ATTACHMENT0,
					       GL_TEXTURE_2D,
					       textures[0],
					       0 /* level */);
			assert(glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER)
			       == GL_FRAMEBUFFER_COMPLETE);

			glDrawArrays(GL_QUADS, 0, 4);

			/* Begin pass 2:
			 */
			glUseProgram(pass[1]);

			glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER,
					       GL_COLOR_ATTACHMENT0,
					       GL_TEXTURE_2D,
					       textures[1],
					       0 /* level */);
			assert(glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER)
			       == GL_FRAMEBUFFER_COMPLETE);

			glDrawArrays(GL_QUADS, 0, 4);

			/* Begin pass 3:
			 */
			glUseProgram(pass[2]);

			glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER,
					       GL_COLOR_ATTACHMENT0,
					       GL_TEXTURE_2D,
					       textures[0],
					       0 /* level */);
			assert(glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER)
			       == GL_FRAMEBUFFER_COMPLETE);

			glDrawArrays(GL_QUADS, 0, 4);

			/* Begin pass 4:
			 */
			glEnable(GL_BLEND);
			glUseProgram(pass[3]);

			glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER,
					       GL_COLOR_ATTACHMENT0,
					       GL_TEXTURE_2D,
					       textures[2],
					       0 /* level */);
			assert(glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER)
			       == GL_FRAMEBUFFER_COMPLETE);

			glUniform2f(q_loc[3],
				    texel_width * float(x),
				    texel_height * float(y));
			glDrawArrays(GL_QUADS, 0, 4);

			/* Re-disable blending.
			 */
			glDisable(GL_BLEND);
		}
	}



	/* Finally normalize the accumulated (w, k * w) data for display.
	 */
	glGenFramebuffers(1, &dst_fbo);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, dst_fbo);
	glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER,
			       GL_COLOR_ATTACHMENT0,
			       GL_TEXTURE_2D,
			       dst_image,
			       0 /* level */);

	glUseProgram(normalize_color_kernel);
	const GLint loc = glGetUniformLocation(normalize_color_kernel, "s");
	assert(loc >= 0);

	glUniform1i(loc, 3);

	glDrawArrays(GL_QUADS, 0, 4);

	/* Emit timing information for the curious. :)
	 */
	glEndQuery(GL_TIME_ELAPSED);

	GLuint64 result;
	glGetQueryObjectui64v(q,
			      GL_QUERY_RESULT,
			      &result);
	printf("%fms per pixel (%ld.%ldms total)\n",
	       double(result) / (1000000. * width * height),
	       result / 1000000,
	       (result / 10000) % 100);

	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);

	glDeleteFramebuffers(1, &tmp_fbo);
	glDeleteFramebuffers(1, &dst_fbo);
	glDeleteTextures(ARRAY_SIZE(textures), textures);
	glDeleteQueries(1, &q);
}

static void
Reshape(int width, int height)
{
	my_surf = SDL_SetVideoMode(width, height, 0,
				   SDL_OPENGL | SDL_RESIZABLE);
	if (my_surf == NULL) {
		exit(1);
	}

	glewInit();
	glViewport(0, 0, width, height);
}

static void
Key(SDLKey sym, bool key_down)
{
	if (!key_down)
		return;

	switch (sym) {
	case SDLK_ESCAPE:
		done = true;
		break;

	case 'o':
		what_to_show = image_original;
		draw = true;
		break;

	case 'n':
		what_to_show = image_denoised;
		draw = true;
		break;

	case 'd':
		what_to_show = image_difference;
		draw = true;
		break;

	default:
		break;
	}
}

void
generate_array_objects()
{
	GLuint vao;
	GLuint vbo[2];

	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	glGenBuffers(ARRAY_SIZE(vbo), vbo);

	static const float data[] = {
		/* Position data */
		-1., -1.,
		 1., -1.,
		 1.,  1.,
		-1.,  1.,

		/* Texture coordinate data */
		0., 0.,
		1., 0.,
		1., 1.,
		0., 1.,
	};

	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STATIC_DRAW);
	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(float),
			      BUFFER_OFFSET(0));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(float),
			      BUFFER_OFFSET(4 * 2 * sizeof(float)));
	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
}

GLint
compile(const char *filename, GLenum target)
{
	gluInitializeCompiler();

	const char *target_name;
	switch (target) {
	case GL_VERTEX_SHADER:
		target_name = "vertex shader";
		break;
	case GL_FRAGMENT_SHADER:
		target_name = "fragment shader";
		break;
	case GL_GEOMETRY_SHADER:
		target_name = "geometry shader";
		break;
	default:
		target_name = "unknown shader";
		break;
	}

	const GLchar *code = gluLoadTextFile(filename);
	if (code == NULL) {
		fprintf(stderr, "could not load %s\n", filename);
		exit(1);
	}

	char *compile_log = NULL;
	GLint sh = gluCompileShader(target, code, &compile_log);

	gluUnloadTextFile(code);

	if (compile_log != NULL) {
		printf("%s log %s:\n%s\n", target_name, filename, compile_log);
		gluReleaseInfoLog(compile_log);
	}

	if (sh == 0)
		exit(1);

	return sh;
}

int
main(int argc, char **argv)
{
	(void) argc;
	(void) argv;

	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		exit(1);
	}

	atexit(SDL_Quit);

	Reshape(1020, 680);

	glGenTextures(ARRAY_SIZE(images), images);
	glBindTexture(GL_TEXTURE_2D, images[0]);
	texture_image(GL_TEXTURE_2D, 0, GL_RGBA,
		      argv[1], false);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,
				 0,
				 GL_TEXTURE_WIDTH,
				 &w);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,
				 0,
				 GL_TEXTURE_HEIGHT,
				 &h);

	glBindTexture(GL_TEXTURE_2D, images[1]);
	glTexImage2D(GL_TEXTURE_2D,
		     0 /* level */,
		     GL_RGBA8,
		     w, h, 0 /* border */,
		     GL_RGBA, GL_UNSIGNED_BYTE, NULL /* data */);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);


	glGenFramebuffers(1, &fbo);
	glBindFramebuffer(GL_READ_FRAMEBUFFER, fbo);
	glFramebufferTexture2D(GL_READ_FRAMEBUFFER,
			       GL_COLOR_ATTACHMENT0,
			       GL_TEXTURE_2D,
			       images[0],
			       0 /* level */);

	static const struct {
		const char *vs;
		const char *fs;
	} shaders[] = {
		{
			"passthrough.vert",
			"difference-kernel.frag"
		},
		{
			"weight-general.vert",
			"weight-5x5-general.frag"
		},
		{
			"passthrough.vert",
			"normalize-kernel.frag"
		},
		{
			"weight-general.vert",
			"weight-5x5-0,0.frag"
		},
		{
			"weight-general.vert",
			"weight-separable-pass1.frag"
		},
		{
			"weight-general.vert",
			"weight-separable-5x5-pass2.frag"
		},
		{
			"weight-general.vert",
			"weight-separable-5x5-pass3.frag"
		},
		{
			"weight-general.vert",
			"weight-separable-5x5-pass4.frag"
		},
		{
			"passthrough.vert",
			"normalize-color-kernel.frag"
		},
	};

	GLuint progs[ARRAY_SIZE(shaders)];

	for (unsigned i = 0; i < ARRAY_SIZE(shaders); i++) {
		GLint fs = compile(shaders[i].fs, GL_FRAGMENT_SHADER);
		GLint vs = compile(shaders[i].vs, GL_VERTEX_SHADER);

		progs[i] = glCreateProgram();
		gluAttachShaders(progs[i], vs, fs, 0);

		char *log = NULL;
		bool is_linked = gluLinkProgram(progs[i], &log);

		if (log != NULL) {
			printf("link log (%s/%s):\n%s\n",
			       shaders[i].vs,
			       shaders[i].fs,
			       log);
			gluReleaseInfoLog(log);
			log = NULL;
		}

		if (!is_linked) {
			exit(1);
		}
	}

	difference_kernel = progs[0];
	alt_weight_kernel = progs[1];
	normalize_kernel = progs[2];
	alt_weight_kernel00 = progs[3];
	pass[0] = progs[4];
	pass[1] = progs[5];
	pass[2] = progs[6];
	pass[3] = progs[7];
	normalize_color_kernel = progs[8];

	generate_array_objects();

	while (!done) {
		SDL_Event event;

		SDL_WaitEvent(&event);
		do {
			switch (event.type) {
			case SDL_VIDEORESIZE:
				Reshape(event.resize.w, event.resize.h);
				draw = true;
				break;

			case SDL_QUIT:
				done = true;
				break;

			case SDL_KEYDOWN:
				Key(event.key.keysym.sym, true);
				break;

			case SDL_KEYUP:
				Key(event.key.keysym.sym, false);
				break;

			case SDL_MOUSEBUTTONUP:
				break;

			default:
				break;
			}
		} while (SDL_PollEvent(&event));

		Redisplay();
	}

	SDL_Quit();
	return 0;
}
